/* data structures and globals for RNPL */
/* $Header: /home/felipe/rmarsa/source/rnpl/RCS/rnpl.h,v 2.6 1996/08/06 20:18:28 rmarsa Exp $ */
/* Copyright (c) 1994-1996 by Robert L. Marsa */

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <bbh_types.h>

#define IVEL 4

/* typedefs */

struct dp;
struct gfc;
struct nml;
struct fn;
struct cd;
struct gdt;
struct ifsta;

typedef struct stl {
  BBH_I type;
  BBH_C *str;
  struct stl *next;
} statement_list;

typedef struct {
  BBH_I type;
  union {
    BBH_C *str;
    BBH_I inum;
    BBH_D num;
  } data;
} scalar;

typedef struct sl {
  scalar sc;
  struct sl *next;
} scalar_list;

typedef struct nd {
  BBH_I type;
  union {
    BBH_C op;
    BBH_D num;
    BBH_I inum;
    struct nml *name;
    struct dp *deriv;
    struct fn *func;
    struct gfc *gfunc;
    struct cd *coord;
  } data;
  struct nd *left;
  struct nd *right;
} node;

typedef struct {
	BBH_I type;
	union {
		BBH_I offset;
		node *expr;
	}val;
} indel;

typedef struct indl {
	indel *ind;
	struct indl *next;
} index_list;

typedef struct nml{
  BBH_C *n;
  struct nml *next;
} name_list;

typedef struct {
	BBH_I ato;
	BBH_I num;
	node *expr;
} work_array;

typedef struct wl {
	work_array *work;
	struct wl *next;
} work_list;

typedef union {
  BBH_C **s_ar;
  BBH_I *i_ar;
  BBH_D *f_ar;
} vector;

typedef struct ivl {
	BBH_I low,high,mod;
	struct ivl *next;
} ivec_list;

typedef struct pd {
  BBH_I type;
  BBH_I size;
	BBH_I con;
  name_list *name;
  BBH_I def;
  vector def_val;
  struct pd *next;
} param_dec;

typedef struct {
  BBH_I type;
  BBH_I size;
	BBH_I con;
  name_list *name;
  BBH_I def;
  vector def_val;
} param_table;

typedef struct prl {
	param_table *par;
	struct prl *next;
} param_ref_list;

typedef struct ad {
  BBH_I type;
  BBH_I encoding;
  name_list *name;
  BBH_I def;
  scalar_list *def_val;
  struct ad *next;
} attrib_dec;

typedef struct {
  BBH_I type;
  BBH_I size;
  BBH_I encoding;
  name_list *name;
  BBH_I def;
  vector def_val;
} attrib_table;

typedef struct cd {
  name_list *name;
  struct gdt *grd;
  indel *indx;
} coord_ref;

typedef struct crd {
  name_list *name;
  struct crd *next;
} coord_list;

typedef struct {
	name_list *name;
	coord_list *subs;
} name_ref;

typedef struct refl{
	BBH_I type;
	union {
		name_ref *nref;
		work_array *work;
	} data;
	struct refl *next;
} ref_list;

typedef struct crdd {
  BBH_I rank;
  coord_list *c_names;
	name_list *name;
	struct crdd *next;
} coord_dec;

typedef struct {
	BBH_I rank;
	coord_list *c_names;
	name_list *name;
} coord_table;

typedef struct ir {
  node *lower;
  node *upper;
  BBH_I inc;
  struct ir *next;
} i_reg;

typedef struct cr {
  name_list *lower;
  name_list *upper;
  struct cr *next;
} c_reg;

typedef struct {
  BBH_I rank;
  i_reg *bounds;
  c_reg *limits;
} region;

typedef struct tof {
  BBH_I offset;
  struct tof *next;
} offset_type;

typedef struct gd {
  BBH_I type;
	name_list *cname;
  name_list *name;
	coord_list *clst;
  i_reg *ireg;
  c_reg *creg;
  struct gd *next;
} grid_list;

typedef struct gdt{
  BBH_I type;
	coord_table *crds;
  name_list *name;
	coord_list *clst;
  region reg;
} grid_table;

typedef struct gfc {
  BBH_I toff;
  name_list *name;
  index_list *indx;
  struct gdt *grd;
} gfunction;

typedef struct fn {
  name_list *name;
  node *expr;
} func;

typedef struct dp{
  BBH_I expand;
  name_list *name;
  node *expr;
  coord_table *clst;
} d_op;

typedef struct op {
  d_op *deriv;
  node *op_expr;
  struct op *next;
} operator;

typedef struct bds{
  BBH_I upper;
  BBH_I lower;
  struct bds *next;
} bounds;

typedef struct {
  name_list *name;
  node *expr;
  coord_table *clst;
  node *op_expr;
  BBH_I ntlevs;
  offset_type *tlev;
  BBH_I ndim;
  bounds *indx;
} dop_table;

typedef struct fd {
  BBH_I type;
	BBH_I alias;
  name_list *fname;
  name_list *gname;
  offset_type *tlev;
  BBH_C *desc;
  struct fd *next;
} gfunc_list;

typedef struct {
  BBH_I type;
	BBH_I alias;
  name_list *fname;
  grid_table *grd;
  BBH_I ntlevs;
  offset_type *tlev;
  BBH_C *desc;
} gfunc_table;

typedef struct gfr {
  gfunc_table *gfunc;
  BBH_I toff;
  BBH_I array;
	name_list *name;
  struct gfr *next;
} gfunc_ref_list;

typedef struct gftbl {
  gfunc_table *gfunc;
	struct gftbl *next;
} gfunc_tab_list;

typedef struct drl {
  dop_table *dop;
  gfunc_ref_list *gf_refs;
  grid_table *grd;
  struct drl *next;
} dop_ref_list;

typedef union {
	node *expr;
	struct ifsta *ifst;
} elstat;

typedef struct ifsta {
	node *lexpr;
	node *expr;
	BBH_I etype; /* (0, 1, 2) -> (none, expr, ifstat) */
	elstat elst;
} ifstat;

typedef struct rlst {
	i_reg *reg;
	BBH_I etype;
	elstat elst;
	struct rlst *next;
} res_list;

typedef struct rd {
	BBH_I eval;
  BBH_I toff;
  offset_type *indx;
  name_list *name;
	res_list *res;
  struct rd *next;
} residual;

typedef struct {
  BBH_I output;
  BBH_I eval;
  BBH_I toff;
  offset_type *indx;
  gfunc_table *gfunc;
  i_reg *reg;
	BBH_I etype;
	elstat elst;
  dop_ref_list *dop_refs;
  gfunc_ref_list *gf_refs;
  gfunc_ref_list *glob_gf;
	coord_list *crds;
	coord_list *glob_crds;
	coord_list *cdifs;
	coord_list *glob_cdifs;
	param_ref_list *parms;
	param_ref_list *glob_par;
} res_table;

typedef struct upd {
  name_list *type;
  name_list *name;
  coord_list *gfs;
  ref_list *refs;
  struct upd *next;
} update;

typedef struct {
	BBH_I header;
  gfunc_tab_list *gfs;
  name_list *type;
  name_list *name;
  gfunc_ref_list *glob_gf;
	param_ref_list *glob_par;
	coord_list *glob_crds;
	coord_list *glob_cdifs;
	work_list *work_refs;
} update_table;

typedef struct id {
  name_list *name;
  res_list *init;
  struct id *next;
} init_dec;

typedef struct {
  gfunc_table *gfunc;
  i_reg *reg;
	BBH_I etype;
	elstat elst;
} init_table;

typedef struct qn {
  node *expr;
  struct qn *next;
} qnode;

typedef struct {
  qnode *head;
  qnode *tail;
} qheader;

#define TMPSIZE 2048
#define TRUE 1
#define FALSE 0

BBH_V fatal_error(const BBH_C *msg);
BBH_I fort_out_const(BBH_C *s, const BBH_D d);
BBH_V fort_out(FILE *fp, const BBH_C *s);

BBH_I strcmp_nc(const BBH_C *s1, const BBH_C *s2);
BBH_I mystrcmp(const BBH_C *s1, const BBH_C *s2);
BBH_V dump_name(const BBH_C *s, const name_list *nm);
name_list *add_name(const BBH_C *nm);
name_list *name_lookup(const BBH_C *nm);

ivec_list *new_ivec_list();
param_ref_list *new_param_ref();
index_list *new_index_list();
indel *new_indel();
work_list *new_work_list();
work_array *new_work_array();
name_ref *new_name_ref();
ref_list *new_ref_list();
statement_list *new_statement_list();
scalar_list *new_scalar_list();
param_dec *new_param_dec();
attrib_dec *new_attrib_dec();
init_dec *new_init_dec();
BBH_I *new_i_ar(const BBH_I i);
BBH_D *new_f_ar(const BBH_I i);
BBH_C **new_s_ar(const BBH_I i);
BBH_C *new_string(const BBH_I i);
name_list *new_name();
coord_ref *new_coord();
gfunction *new_gfunc();
d_op *new_dop();
node *new_node();
offset_type *new_offset();
func *new_func();
coord_list *new_coord_list();
coord_dec *new_coord_dec();
coord_table *new_coord_tab();
update *new_update();
bounds *new_bounds();
dop_ref_list *new_dop_ref();
gfunc_ref_list *new_gfunc_ref();
gfunc_tab_list *new_gfunc_tab();
grid_list *new_grid_list();
gfunc_list *new_gfunc_list();
operator *new_operator();
res_list *new_res_list();
residual *new_residual();
i_reg *new_i_reg();
c_reg *new_c_reg();
ifstat *new_ifstat();

BBH_V gfunc_suffix(FILE *fp,const BBH_I offset);
BBH_I get_priority(const BBH_C c);
BBH_V do_left_paren(FILE *fp, node *expr, BBH_C *c);
BBH_V do_right_paren(FILE *fp, node *expr, BBH_C *c);
BBH_V ex_to_string(FILE *fp,node *expr);
BBH_V dop_to_string(FILE *fp,d_op *d);
BBH_V func_to_string(FILE *fp,func *f);
BBH_V coord_to_string(FILE *fp,coord_ref *cd);
BBH_V gfunc_to_string(FILE *fp,gfunction *f);
BBH_V array_bounds_to_string(FILE *fp, const grid_table *gr, const BBH_I ar);
BBH_V array_reff0_to_string(FILE *fp,const grid_table *gr);
BBH_V array_reff_to_string(FILE *fp, const gfunction *f, const grid_table *gr);
BBH_V array_ref0_to_string(FILE *fp,const grid_table *gr);
BBH_V array_ref_to_string(FILE *fp,const gfunction *f,const grid_table *gr);
BBH_V grid_size(FILE *fp,const grid_table *gr);

BBH_V show_name_type(const name_list *nm);
BBH_I name_type(const name_list *nm);
BBH_I is_upfile(const name_list *nm);
BBH_I is_update(const name_list *nm);
BBH_I is_func(const name_list *nm);
BBH_I is_dop(const name_list *nm);
BBH_I is_attrib(const name_list *nm);
BBH_I is_coord_sys(const name_list *nm);
BBH_I is_grid(const name_list *nm);
BBH_I is_cdif(const name_list *nm);
BBH_I is_gbase(const name_list *nm);
BBH_I is_param(const name_list *nm);
BBH_I is_const(const name_list *nm);
gfunc_table *name_to_gfunc(const name_list *nm);
dop_table *find_dop(const name_list *nm, const coord_table *cl);
BBH_I resid_exists(const name_list *nm);
BBH_I in_clst(const name_list *nm, const coord_table *c1);
BBH_I is_coord(const name_list *nm);
BBH_I is_space(const name_list *nm);
BBH_I rank(const name_list *nm);
BBH_I compare_ireg(const i_reg *i1, const i_reg *i2);
BBH_I compare_ctab(const coord_table *c1, const coord_table *c2);
BBH_I compare_offset(const offset_type *i1, const offset_type *i2);
BBH_I compare_index(const index_list *i1, const index_list *i2);
BBH_I compare_expr(const node *e1, const node *e2);
BBH_V expand_expr(node *expr);
BBH_V expand_ifstat(ifstat *ifst);
BBH_V evaluate_expr(node *expr);
BBH_V evaluate_ifstat(ifstat *ifst);
BBH_V offset_expr(const BBH_I toff, offset_type *ind, node *expr);
BBH_V offset_ifstat(const BBH_I toff, offset_type *ind, ifstat *ifst);
BBH_V simplify_expr(node *expr);
BBH_V simplify_ifstat(ifstat *ifst);
BBH_V make_expand(node *expr);
BBH_I check_expr(node *expr);
BBH_I check_ifstat(ifstat *ifst);
node *deriv_wrt_gfunc(node *expr, node *gf);
BBH_V eval_wrt_gfunc(node *expr, node *gf);
BBH_V del_clst(coord_list *cl);
coord_list *dup_clst(coord_list *cl);
coord_table *dup_ctab(coord_table *cl);
gfunc_ref_list *dup_gf_ref(gfunc_ref_list *grl);
param_ref_list *dup_param_ref(param_ref_list *prl);
indel *dup_indel(indel *indx);
index_list *dup_indx(index_list *indx);
node *dup_expr(node *expr);
BBH_V del_expr(node *expr);
grid_table *get_grid(node *ex);
BBH_V set_grid(grid_table *grd, node *ex);
BBH_V name_subst(name_list *nm, grid_table *grd, node *ex, const name_list *s_nm);
BBH_V coord_subst(coord_ref *crd, node *ex);
BBH_V gfunc_subst(gfunction *gf, node *ex, const name_list *s_nm);
BBH_V expr_subst(node *e1, node *ex, const name_list *s_nm);
BBH_V gfunc_stuff(node *e1, node *ex);
qheader *init_queue();
BBH_V enqueue(qheader *qh, node *expr);
node *dequeue(qheader *qh);
BBH_V add_offset(offset_type *l, offset_type *o);
BBH_V add_dop_ref(dop_ref_list *l, dop_ref_list *o);
BBH_V add_gfunc_ref(gfunc_ref_list *l, gfunc_ref_list *o);
BBH_V add_name_ref(coord_list *l, coord_list *o);
BBH_V add_param_ref(param_ref_list *l, param_ref_list *o);
BBH_V add_param(param_dec *l, param_dec *o);
BBH_V add_attrib(attrib_dec *l, attrib_dec *o);
BBH_V add_work_ref(work_list *l, work_list *o);
offset_type *find_tlevs(const node *n);
bounds *find_indx(const node *n);
gfunc_ref_list *find_gfs(const node *ex);
dop_ref_list *find_dops(const node *ex);
dop_ref_list *find_dops_if(const ifstat *ifst);
gfunc_ref_list *find_gfuncs(const node *ex);
gfunc_ref_list *find_gfuncs_if(const ifstat *ifst);
coord_list *find_coords(const node *ex);
coord_list *find_coords_if(const ifstat *ifst);
coord_list *find_cdifs(const node *ex);
coord_list *find_cdifs_if(const ifstat *ifst);
param_ref_list *find_params(const node *ex);
param_ref_list *find_params_if(const ifstat *ifst);
BBH_I get_size_one();
BBH_I get_size_all();
BBH_I gptr_to_index(const grid_table *gr);
BBH_I update_to_index(const update_table *up);
BBH_I gfunc_to_index(const gfunc_table *gf);
BBH_I offset_to_index(const BBH_I tof, offset_type *ofl);
BBH_I check_gen_expr(node *ex);
BBH_I check_gen_ifstat(ifstat *ifst);
BBH_I check_simp_expr(node *ex);
BBH_I check_log_expr(node *ex);

BBH_V resid_header(FILE *fp, const res_table *rs);
BBH_V resid_call(FILE *fp, const res_table *rs);
BBH_V update_header(FILE *fp, const update_table *up);
BBH_V update_call(FILE *fp, const update_table *up);
BBH_V output_resid_stat(FILE *fp, const res_table *rs);
BBH_V write_single_stat(FILE *fp, const res_table *rs, node *e);
BBH_V output_update_stat(FILE *fp, const res_table *rs);
BBH_V declare_coord_difs(FILE *fp);
BBH_V declare_parameters(FILE *fp);
BBH_V declare_grids(FILE *fp);
BBH_V declare_gfuncs(FILE *fp);
BBH_V declare_attributes(FILE *fp);
BBH_V code_residuals(FILE *fp);
BBH_V code_residuals_f77(FILE *fp);
BBH_V code_residuals_c(FILE *fp);
BBH_V swap_top(FILE *fp);
BBH_V swap_top_f77(FILE *fp);
BBH_V swap_top_c(FILE *fp);
BBH_V swap_levels(FILE *fp);
BBH_V swap_levels_f77(FILE *fp);
BBH_V swap_levels_c(FILE *fp);
BBH_V one_step(FILE *fp);
BBH_V one_step_f77(FILE *fp);
BBH_V one_step_c(FILE *fp);
BBH_V update_helpers(FILE *fp);
BBH_V update_helpers_f77(FILE *fp);
BBH_V update_helpers_c(FILE *fp);
BBH_V code_updates(FILE *fp);
BBH_V init_params_attribs(FILE *fp);
BBH_V init_params_attribs_f77(FILE *fp);
BBH_V init_params_attribs_c(FILE *fp);
BBH_V init_coord_difs(FILE *fp);
BBH_V init_coord_difs_f77(FILE *fp);
BBH_V init_coord_difs_c(FILE *fp);
BBH_V init_grids(FILE *fp);
BBH_V init_grids_f77(FILE *fp);
BBH_V init_grids_c(FILE *fp);
BBH_V init_gfuncs(FILE *fp);
BBH_V init_gfuncs_f77(FILE *fp);
BBH_V init_gfuncs_c(FILE *fp);
BBH_V read_gfuncs(FILE *fp,BBH_C *iname);
BBH_V read_gfuncs_f77(FILE *fp,BBH_C *iname);
BBH_V read_gfuncs_c(FILE *fp,BBH_C *iname);
BBH_V write_attributes(FILE *fp);
BBH_V write_attributes_f77(FILE *fp);
BBH_V write_attributes_c(FILE *fp);
BBH_V init_fragment(FILE *fp);
BBH_V cleanup(FILE *fp);
BBH_V handler(FILE *fp);
BBH_V handler_f77(FILE *fp);
BBH_V handler_c(FILE *fp);
BBH_V dump_state(FILE *fp);
BBH_V dump_state_f77(FILE *fp);
BBH_V dump_state_c(FILE *fp);
BBH_V gen_gfuncs(FILE *fp);
BBH_V gen_gfuncs_f77(FILE *fp);
BBH_V gen_gfuncs_c(FILE *fp);
BBH_V code_iteration(FILE *fp);
BBH_V code_iteration_f77(FILE *fp);
BBH_V code_iteration_c(FILE *fp);

#define FLAG 100000

#ifdef _SUPPORT_

/* globals */
name_list *Ln,*Sin,*Cos,*Exp,*Tan,*Sqrt,*Sinh,*Cosh,*Tanh;
BBH_I ncoords; /* number of coordinate systems */
BBH_I nparams; /* number of parameters */
BBH_I ngrids; /* number of grids */
BBH_I ngfuncs; /* number of functions */
BBH_I nopers; /* number of derivative operators */
BBH_I nresids; /* number of residuals */
BBH_I language; /* output language */
BBH_I nupdates; /* number of updates */
BBH_I nattribs; /* number of attributes */
BBH_I ninits; /* number of initializations */
BBH_C ldir[70]; /* lib directory */

coord_table *coords; /* table of coordinate systems */
name_list ***coord_difs; /* table of coordinate differentials */
name_list ***grid_base; /* table of grid bases */
param_table *params; /* table of parameters */
grid_table *grids; /* table of grids */
gfunc_table *gfuncs; /* table of functions defined on the grids */
attrib_table *attribs; /* table of attributes */
dop_table *dopers; /* table of derivative operators */
res_table *resids; /* table of residuals */
update_table *updates; /* table of updates */
init_table *inits; /* table of initializations */
name_list str_tbl; /* list of strings */
statement_list *stmts; /* list of statements */
name_list *loop_driver; /* name of driver file */
work_list *static_work; /* list of static work arrays */

BBH_I fort_const_D=1,adapt=0,parallel=0;
#else

/* globals */
extern name_list *Ln,*Sin,*Cos,*Exp,*Tan,*Sqrt,*Sinh,*Cosh,*Tanh;
extern BBH_I ncoords; /* number of coordinate systems */
extern BBH_I nparams; /* number of parameters */
extern BBH_I ngrids; /* number of grids */
extern BBH_I ngfuncs; /* number of functions */
extern BBH_I nopers; /* number of derivative operators */
extern BBH_I nresids; /* number of residuals */
extern BBH_I language; /* output language */
extern BBH_I nupdates; /* number of updates */
extern BBH_I nattribs; /* number of attributes */
extern BBH_I ninits; /* number of initializations */
extern BBH_C ldir[70]; /* lib directory */

extern coord_table *coords; /* table of coordinate systems */
extern name_list ***coord_difs; /* table of coordinate differentials */
extern name_list ***grid_base; /* table of grid bases */
extern param_table *params; /* table of parameters */
extern grid_table *grids; /* table of grids */
extern gfunc_table *gfuncs; /* table of functions defined on the grids */
extern attrib_table *attribs; /* table of attributes */
extern dop_table *dopers; /* table of derivative operators */
extern res_table *resids; /* table of residuals */
extern update_table *updates; /* table of updates */
extern init_table *inits; /* table of initializations */
extern name_list str_tbl; /* list of strings */
extern statement_list *stmts; /* list of statements */
extern name_list *loop_driver; /* name of driver file */
extern work_list *static_work; /* list of static work arrays */

extern BBH_I fort_const_D,adapt,parallel;
#endif
